/*
 *  tcpjunk - TCP protocols testing tool
 *  Contact: oferkv@gmail.com
 *
 *  tcpjunk is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by the
 *  Free Software Foundation; either version 2, or (at your option) any
 *  later version.
 *
 *  tcpjunk is distributed in the hope that it will be useful, but WITHOUT
 *  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 *  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 *  for more details.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/wait.h>
#include <netdb.h>
#include <sys/ioctl.h>
#include <net/if.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <fcntl.h>
#include <dirent.h>
#include "tcpjunk.h"

void show_help()
  {
  printf("tcpjunk v%s - TCP protocols testing tool\n", VER);
  puts("\nAct as a client:");
  puts(" tcpjunk -s <hostname> -p <port> -c <session file> [options]");
  puts("\nGUI mode:");
  puts(" tcpjunk -x");
  puts("\nAct as a server:");
  puts(" tcpjunk -r -p <port> -c <session file> [options]");
  puts("\nOptions:");
  puts("  -c <filename>  session setup file");
  puts("  -f             use random session files from current directory");
  puts("  -s <hostname>  hostname/IP of target");
  puts("  -r             server mode");
  puts("  -p <port>      TCP port to connect/listen to");
  puts("  -t <sec>       initial timeout (default: 2 seconds)");
  puts("  -T <ms>        reply timeout (default: 100 milliseconds)");
  puts("  -W             wait for data at first");
  puts("  -C             dont wait for data at last");
  puts("  -m <bytes>     max segment size");
  puts("  -D <ms>        delay before sending each segment");
  puts("  -P <IP>        use specific source IP address");
  puts("  -R             use random source IP addresses");
  puts("  -I <interface> source IP alias interface");
  puts("  -o <offset>    source IP interface alias offset");
  puts("  -u             optimize TCP tw_recycle, tw_reuse settings and");
  puts("                 file descriptor limits");
  puts("  -i <number>    number of instances to run in parallel");
  puts("  -l <number>    number of loops for each instance (default: forever)");
  puts("  -w <mseconds>  delay between connections (default: none)");
  puts("  -k <times>     repeat on the same connection (default: 1)");
  puts("  -S             use SSL v3.0");
  puts("  -E <filename>  server Certificate file (default: server.crt)");
  puts("  -K <filename>  server key file (default: server.key)");
  puts("  -d             force display of session data when using instances");
  puts("                 or in server mode");
  puts("  -v             display warnings and other information");
  puts("  -h             show this help and exit");
  puts("\nSession setup file tags:");
  puts(" <section>[data]</section>           delimiters for multiple sections");
  puts(" <repeat [min] [max]>[data]</repeat> data repeater");
  puts(" <system>[command]<system>           insert output of a command");
  puts(" <fuzz [type] [min] [max]>           insert a fuzzed buffer");
  puts(" <char [hex code]>                   insert a character");
  puts(" <instance>                          insert client instance number");
  puts("\nFuzzing types:");
  puts(" digit    [0-9]");
  puts(" alnum    [0-9][a-z][A-Z]");
  puts(" alpha    [a-z][A-Z]");
  puts(" cntrl    control characters");
  puts(" print    any printable character");
  puts(" special  above ascii 127");
  puts(" punct    punctuation characters");
  puts(" space    space characters");
  puts(" any      any character");
  puts(" random   random fuzzing type for each instance");
  puts("\nBug reports and suggestions: <oferkv@gmail.com>");

  exit(EXIT_FAILURE);
  }

int read_reply(int sock, SSL *ssl, struct tcpjunk_data *instance_data)
  {
  int sel_ret, read_ret;
  int read_success = 0;
  char read_buf[SOCK_BUF_LEN];

  /* initialize descriptors set */
  fd_set set;
  FD_ZERO (&set);
  FD_SET (sock, &set);

  /* initialize timeout */
  struct timeval timeout;
  timeout.tv_sec = instance_data->itimeout;
  timeout.tv_usec = 0;

  while ((sel_ret = select(sock + 1, &set, NULL, NULL, &timeout))> 0
  || (instance_data->use_ssl && SSL_pending(ssl)))
    {
    /* clean buffer */
    memset (read_buf, 0, sizeof (read_buf));

    /* read clear text */
    if (instance_data->use_ssl == 0)
      {
      read_ret = recv (sock, read_buf, SOCK_BUF_LEN, MSG_DONTWAIT);
      if (read_ret < 0)
        {
        tj_error (instance_data, "Error reading data from socket: ",strerror(errno));
        break;
        }
      }
    /* read using SSL */
    else
      {
      read_ret = SSL_read (ssl, read_buf, SOCK_BUF_LEN);
      if (read_ret < 0)
        {
        tj_error (instance_data, "SSL read ", ERR_error_string(ERR_get_error(), NULL));
        break;
        }
      }

    /* if no data break */
    if (read_ret == 0)
    break;

    /* print reply */
    if (instance_data->debug> 0 && instance_data->gui == 0)
    fwrite(read_buf, sizeof (char), read_ret, stdout);
    else
    if (instance_data->gui == 1)
    gui_show_traffic (read_buf, read_ret, 1);

    if (instance_data->verbose)
      {
      char read_len_str[64];
      sprintf(read_len_str, "received %d bytes", read_ret);
      tj_error (instance_data, read_len_str, "");
      }

    if (!read_success)
    read_success = 1;

    /* reset timeouts */
    timeout.tv_sec = 0;
    timeout.tv_usec = instance_data->stimeout;

    }

  if (read_success)
  return 0;
  else
  return 1;

  }

void send_data(int sock, SSL *ssl, struct tcpjunk_data *instance_data,
    char *send_buf, int send_buf_len)
  {

  int rc;
  int err;

  if (instance_data->use_ssl == 0)
    {

    /* send junk clear text */
    if (instance_data->segment_size == 0)
      {
      rc = send(sock, send_buf, send_buf_len, 0);
      if (rc < 0)
        tj_error(instance_data, "Error sending data: ", strerror(errno));
      }
    else
      {
      /* segment size */
      int loc, seg_len = instance_data->segment_size;
      for (loc = 0; loc < send_buf_len; loc += seg_len)
        {
        if ((send_buf_len - loc) < seg_len)
          seg_len = send_buf_len - loc;
        rc = send(sock, &send_buf[loc], seg_len, 0);
        if (instance_data->segment_delay > 0)
          usleep(instance_data->segment_delay);
        if (rc < 0)
          {
          tj_error(instance_data, "Error sending data: ", strerror(errno));
          break;
          }
        }
      }
    }
  else
    {
    /* send junk using SSL */

    if (instance_data->segment_size == 0)
      {
      err = SSL_write(ssl, send_buf, send_buf_len);
      if ((err) == -1)
        tj_error(instance_data, "SSL write ", ERR_error_string(ERR_get_error(),
            NULL));
      }
    else
      {
      /* segment size */
      int loc, seg_len = instance_data->segment_size;
      for (loc = 0; loc < send_buf_len; loc += seg_len)
        {
        if ((send_buf_len - loc) < seg_len)
          seg_len = send_buf_len - loc;
        err = SSL_write(ssl, &send_buf[loc], seg_len);
        if (instance_data->segment_delay > 0)
          usleep(instance_data->segment_delay);
        if ((err) == -1)
          tj_error(instance_data, "SSL write ", ERR_error_string(
              ERR_get_error(), NULL));
        }
      }
    }

  /* print sent buffer if we are debugging */
  if (instance_data->debug > 0 && instance_data->gui == 0)
    fwrite(send_buf, sizeof(char), send_buf_len, stdout);
  else if (instance_data->gui == 1)
    {
    gui_show_traffic(send_buf, send_buf_len, 0);
    }
  if (instance_data->verbose)
    {
    char read_len_str[64];
    sprintf(read_len_str, "sent %d bytes", send_buf_len);
    tj_error(instance_data, read_len_str, "");
    }

  }

int tj_connect(int sock, struct sockaddr_in sa,
    struct tcpjunk_data *instance_data)
  {
  int ret = 0, sock_opts;

  /* set non-blocking on socket */
  if ((sock_opts = fcntl(sock, F_GETFL, 0)) >= 0)
    ret = fcntl(sock, F_SETFL, sock_opts | O_NONBLOCK);
  if (ret < 0)
    return -1;

  /* connect */
  ret = connect(sock, (struct sockaddr *) &sa, sizeof(sa));
  if ((ret < 0) && (errno!= EINPROGRESS))
    return -1;

  /* initialize descriptors set */
  fd_set set;
  FD_ZERO (&set);
  FD_SET (sock, &set);

  /* initialize timout */
  struct timeval timeout;
  timeout.tv_sec = instance_data->itimeout;
  timeout.tv_usec = 0;

  ret = select(sock + 1, NULL, &set, NULL, &timeout);
  if (ret > 0)
    {
    int so_ret, get_ret;
    unsigned int get_len = sizeof(get_ret);

    /* handle errors */
    so_ret = getsockopt(sock, SOL_SOCKET, SO_ERROR, &get_ret, &get_len);
    if (so_ret < 0)
      {
      tj_error(instance_data, "System request failed: ", strerror(errno));
      return -1;
      }
    if (get_ret > 0)
      {
      errno = get_ret;
      return -1;
      }
    }
  else if (ret <= 0)
    {
    errno = ETIMEDOUT;
    return -1;
    }

  /* all was good, restore socket options */
  ret = fcntl(sock, F_SETFL, sock_opts);

  if (instance_data->verbose == 1)
    tj_error(instance_data, "Connected", "");
  return 0;
  }

void tj_disconnect(int sock, struct tcpjunk_data *instance_data, SSL * ssl,
    SSL_CTX * ctx)
  {

  int err = 0;
  if (instance_data->use_ssl == 1)
    {
    if (ssl != NULL)
      err = SSL_clear(ssl);
    if ((err) == -1)
      tj_error(instance_data, "SSL Clear ", ERR_error_string(ERR_get_error(),
          NULL));
    }

  err = close(sock);
  if (err < 0)
    tj_error(instance_data, "Closing socket failed: ", strerror(errno));
  else if (instance_data->verbose == 1)
    tj_error(instance_data, "Connection closed", "");

  }

char fuzz_char(int ftype)
  {

  int rg = 0;
  int tr = 0;

  /* digit 48-57 */
  if (ftype == r_digit)
    return rand() % 10 + 48;

  /* alnum 48-57 65-90 97-122 */
  else if (ftype == r_alnum)
    {
    rg = rand() % 3;
    if (rg == 0)
      return rand() % 10 + 48;
    else if (rg == 1)
      return rand() % 26 + 65;
    else if (rg == 2)
      return rand() % 26 + 97;
    }

  /* alpha 65-90 97-122 */
  else if (ftype == r_alpha)
    {
    rg = rand() % 2;
    if (rg == 0)
      return rand() % 26 + 65;
    else if (rg == 1)
      return rand() % 26 + 97;
    }

  /* cntrl 1-31 127 */
  if (ftype == r_cntrl)
    {
    tr = rand() % 33 + 1;
    if (tr == 32)
      tr = 127;
    return tr;
    }

  /* print 32-126 */
  if (ftype == r_print)
    return rand() % 95 + 32;

  /* special 128-255 */
  if (ftype == r_special)
    return rand() % 128 + 128;

  /* punct 33-47 58-64 91-96 123-126 */
  else if (ftype == r_punct)
    {
    rg = rand() % 4;
    if (rg == 0)
      return rand() % 15 + 33;
    else if (rg == 1)
      return rand() % 7 + 58;
    else if (rg == 2)
      return rand() % 6 + 91;
    else if (rg == 3)
      return rand() % 4 + 123;
    }

  /* space 9-13 32 */
  if (ftype == r_space)
    {
    tr = rand() % 6 + 9;
    if (tr == 14)
      tr = 32;
    return tr;
    }

  /* any 0-255 */
  if (ftype == r_any)
    return rand() % 255;

  /* we never reach here */
  return 0;
  }

/* check fuzz type */
int get_fuzz_type(char *fuzz_type, struct tcpjunk_data *instance_data)
  {
  if (strcmp(fuzz_type, "none") == 0)
    return r_none;
  else if (strcmp(fuzz_type, "digit") == 0)
    return r_digit;
  else if (strcmp(fuzz_type, "alnum") == 0)
    return r_alnum;
  else if (strcmp(fuzz_type, "alpha") == 0)
    return r_alpha;
  else if (strcmp(fuzz_type, "cntrl") == 0)
    return r_cntrl;
  else if (strcmp(fuzz_type, "print") == 0)
    return r_print;
  else if (strcmp(fuzz_type, "special") == 0)
    return r_special;
  else if (strcmp(fuzz_type, "punct") == 0)
    return r_punct;
  else if (strcmp(fuzz_type, "space") == 0)
    return r_space;
  else if (strcmp(fuzz_type, "any") == 0)
    return r_any;
  else if (strcmp(fuzz_type, "random") == 0)
    return r_random;
  else
    {
    tj_error(instance_data, "Invalid fuzz type - ", fuzz_type);
    if (instance_data->gui == 0)
      exit(EXIT_FAILURE);
    else
      return 0;
    }

  }

void sigchld_handler(int s)
  {
  while (wait(NULL) > 0)
    ;
  }

void load_session_data(struct tcpjunk_data *instance_data)
  {
  /* load session file */
  FILE *fp;
  if ((fp = fopen(instance_data->session_file_name, "r")) == NULL)
    {
    printf("tcpjunk: can not open session file %s\n",
        instance_data->session_file_name);
    exit(EXIT_FAILURE);
    }

  memset(instance_data->session_data, 0, sizeof(SESSION_MAX_LEN));
  int chr, loc = 0;
  while ((chr = fgetc(fp)) != EOF)
    {
    instance_data->session_data[loc] = chr;
    loc++;
    if (loc == SESSION_MAX_LEN)
      {
      printf("tcpjunk: the file %s is too large\n",
          instance_data->session_file_name);
      exit(EXIT_FAILURE);
      }
    }
  instance_data->session_data_len = loc;
  fclose(fp);
  }

int check_address(struct tcpjunk_data *instance_data, char *srv_addr)
  {

  /* check address */
  struct hostent *hnames;
  hnames = gethostbyname(srv_addr);
  if (!hnames)
    {
    tj_error(instance_data, "Invalid server address", "");
    if (instance_data->gui == 0)
      exit(EXIT_FAILURE);
    else
      return 0;
    }
  else
    {
    instance_data->server_addr = hnames->h_addr_list[0];
    instance_data->server_addr_len = hnames->h_length;
    return 1;
    }
  }

int check_port(struct tcpjunk_data *instance_data)
  {
  /* check port */
  if (instance_data->server_port < 1 || instance_data->server_port > 65535)
    {
    tj_error(instance_data, "Invalid port", "");
    if (instance_data->gui == 0)
      exit(EXIT_FAILURE);
    else
      return 0;
    }
  return 1;
  }

void tj_error(struct tcpjunk_data *instance_data, char *message, char *param)
  {
  if (instance_data->debug || instance_data->verbose)
    {
    char temp_err[1024];

    /* print to stderr or in GUI */
    if (instance_data->gui == 0)
      {
      sprintf(temp_err, "%s %s%s\n", "tcpjunk:", message, param);
      fprintf(stderr, "%s", temp_err);
      }
    else
      {
      sprintf(temp_err, "%s%s\n", message, param);
      gui_add_log(temp_err);
      }
    }
  }

void check_session_data(struct tcpjunk_data *instance_data)
  {
  /* find all send buffers in session data */
  char *config_buf = instance_data->session_data;
  instance_data->sect_idx = 0;
  char *section_end;

  /* If there are sections, enumerate them */
  while ((instance_data->sections[instance_data->sect_idx] = strstr(config_buf,
      instance_data->start_section)))
    {
    /* start tag found */
    instance_data->sections[instance_data->sect_idx] += strlen(
        instance_data->start_section);

    /* get section end */
    if (!(section_end = strstr(
        instance_data->sections[instance_data->sect_idx],
        instance_data->end_section)))
      {
      /* end tag not found, using all buffer */
      instance_data->sections[instance_data->sect_idx] = config_buf;
      instance_data->sections_len[instance_data->sect_idx]
          = instance_data->session_data_len;
      if (instance_data->sect_idx == 0)
        instance_data->sect_idx++;
      return;
      }
    /* end tag found */
    instance_data->sections_len[instance_data->sect_idx] = section_end
        - instance_data->sections[instance_data->sect_idx];

    instance_data->sect_idx++;
    config_buf = section_end + 1;
    }

  /* just a single section */
  if (instance_data->sect_idx == 0)
    {
    instance_data->sections[instance_data->sect_idx] = config_buf;
    instance_data->sections_len[instance_data->sect_idx]
        = instance_data->session_data_len;
    instance_data->sect_idx++;
    }

  return;

  }

void set_random_session_file(struct tcpjunk_data *instance_data)
  {
  strcpy(instance_data->session_file_name, instance_data->rand_file_list[rand()
      % instance_data->rand_files_num]);
  }

void load_rand_file_list(struct tcpjunk_data *instance_data)
  {
  int file_index = 0;
  struct dirent* dfile;
  DIR *cur_dir = opendir("./");
  if (!cur_dir)
    {
    perror("tcpjunk: error opening directory ");
    exit(EXIT_FAILURE);
    }
  while (((dfile = readdir(cur_dir)) != NULL) && file_index < MAXFILES)
    {
    if (strcmp(dfile->d_name, "..") != 0 && strcmp(dfile->d_name, ".") != 0)
      {
      strcpy(&instance_data->rand_file_list[file_index][0], dfile->d_name);
      file_index++;
      }
    }
  if (closedir(cur_dir) == -1)
    {
    perror("tcpjunk: error closing directory ");
    exit(EXIT_FAILURE);
    }
  instance_data->rand_files_num = file_index;

  }

void set_ip(char *ifname, char *s_ip)
  {

  /* setup ip on alias */
  in_addr_t ip = inet_addr(s_ip);
  int sock = socket(AF_INET,SOCK_DGRAM, 0);
  struct ifreq ifr;

  strcpy(ifr.ifr_name, ifname);
  ifr.ifr_addr.sa_family = AF_INET;
  ((struct sockaddr_in *) &ifr.ifr_addr)->sin_addr.s_addr = ip;
  if (ioctl(sock, SIOCSIFADDR, &ifr) < 0)
    perror("Error setting up ip address");

  close(sock);

  }

int fuzz_data(char *send_buf, struct tcpjunk_data *instance_data,
    int send_buf_len)
  {
  char *fuzz_tag = "<fuzz ";
  int fuzz_tag_min;
  int fuzz_tag_max;
  char prep_buf[SESSION_MAX_LEN];
  memset(prep_buf, 0, sizeof(prep_buf));
  char *prep_buf_loc = prep_buf;
  char *send_buf_loc = send_buf;
  char tok_fuzz_type[16];
  char *fuzz_tag_loc = NULL;
  char *temp_loc = NULL;
  int tags_processed = 0;
  int fuzz_mode = 0;

  while ((fuzz_tag_loc = strstr(send_buf_loc, fuzz_tag)))
    {
    tags_processed = 1;
    /* add text up to tag */
    memcpy(prep_buf_loc, send_buf_loc, fuzz_tag_loc - send_buf_loc);

    /* advance prep_buf_loc to end */
    prep_buf_loc += fuzz_tag_loc - send_buf_loc;

    /* advance send_buf_loc to end of tag */
    send_buf_loc = fuzz_tag_loc + strlen(fuzz_tag);

    /* parse tag */
    sscanf(send_buf_loc, "%s %d %d", tok_fuzz_type, &fuzz_tag_min,
        &fuzz_tag_max);
    if (!(fuzz_mode = get_fuzz_type(tok_fuzz_type, instance_data)))
      return send_buf_len;
    if (fuzz_tag_min < 1 || fuzz_tag_max < 1 || fuzz_tag_min > fuzz_tag_max)
      return send_buf_len;

    /* advance send_buf_loc to end of tag */
    if (!(send_buf_loc = strstr(send_buf_loc, ">")))
      return send_buf_len;

    send_buf_loc++;
    /* add fuzzed data */
    int rand_buf_len = rand() % (1 + (fuzz_tag_max - fuzz_tag_min))
        + fuzz_tag_min - 1;

    /* set random fuzzing type */
    if (fuzz_mode == r_random)
      fuzz_mode = rand() % 9 + 1;

    for (temp_loc = prep_buf_loc; temp_loc <= prep_buf_loc + rand_buf_len; temp_loc++)
      *temp_loc = fuzz_char(fuzz_mode);
    prep_buf_loc = temp_loc;
    }

  if (tags_processed)
    {
    /* add last part */
    memcpy(prep_buf_loc, send_buf_loc, (send_buf + send_buf_len) - send_buf_loc);

    /* copy new buffer to send buffer */
    memset(send_buf, 0, SESSION_MAX_LEN);
    send_buf_len = (prep_buf_loc - prep_buf + (send_buf + send_buf_len)
        - send_buf_loc);
    memcpy(send_buf, prep_buf, send_buf_len);
    }

  return send_buf_len;
  }

int system_data(char *send_buf, struct tcpjunk_data *instance_data,
    int send_buf_len)
  {
  char *send_buf_loc = send_buf;
  char *system_tag = "<system>";
  char *system_tag_end = "</system>";
  char *system_tag_loc = NULL;
  char prep_buf[SESSION_MAX_LEN];
  char *prep_buf_loc = prep_buf;
  char system_cmd[FILENAME_MAX];
  char *system_cmd_loc;
  int tags_processed = 0;
  FILE *cmd_pipe;
  int pipe_char;

  memset(prep_buf, 0, sizeof(prep_buf));
  memset(system_cmd, 0, sizeof(system_cmd));

  while ((system_tag_loc = strstr(send_buf_loc, system_tag)))
    {
    /* add text up to tag */
    memcpy(prep_buf_loc, send_buf_loc, system_tag_loc - send_buf_loc);

    /* advance prep_buf_loc to end of text */
    prep_buf_loc += system_tag_loc - send_buf_loc;

    /* advance send_buf_loc to end of tag */
    send_buf_loc = system_tag_loc + strlen(system_tag);

    /* point to start of tag data */
    system_cmd_loc = send_buf_loc;

    /* advance send_buf_loc to end of tag data */
    if (!(send_buf_loc = strstr(send_buf_loc, system_tag_end)))
      return send_buf_len;

    /* parse tag */
    strncpy(system_cmd, system_cmd_loc, send_buf_loc - system_cmd_loc);

    /* advance send_buf_loc to end of closing tag */
    send_buf_loc += strlen(system_tag_end);

    /* get stuff from the pipe */
    if ((cmd_pipe = popen(system_cmd, "r")) != NULL)
      {
      while ((pipe_char = getc(cmd_pipe)) != EOF)
        {
        *prep_buf_loc = pipe_char;
        prep_buf_loc++;
        }
      }
    (void) pclose(cmd_pipe);

    tags_processed = 1;
    }

  if (tags_processed)
    {
    /* add last part */
    memcpy(prep_buf_loc, send_buf_loc, (send_buf + send_buf_len) - send_buf_loc);

    /* copy new buffer to send buffer */
    memset(send_buf, 0, SESSION_MAX_LEN);
    send_buf_len = (prep_buf_loc - prep_buf + (send_buf + send_buf_len)
        - send_buf_loc);
    memcpy(send_buf, prep_buf, send_buf_len);
    }
  return send_buf_len;
  }

int repeat_data(char *send_buf, struct tcpjunk_data *instance_data,
    int send_buf_len)
  {
  char *send_buf_loc;
  char *repeat_tag = "<repeat ";
  char *repeat_tag_end = "</repeat>";
  char repeat_text[10240];
  char *repeat_tag_loc = NULL;
  char *repeat_text_loc = NULL;
  int repeat_min;
  int repeat_max;
  char prep_buf[SESSION_MAX_LEN];
  char *prep_buf_loc;
  int tags_processed;
  int nesting;
  char *repeat_tag_nest_loc = NULL;
  int repeat_text_len = 0;

  start:

  nesting = 0;
  tags_processed = 0;
  memset(prep_buf, 0, sizeof(prep_buf));
  memset(repeat_text, 0, sizeof(repeat_text));
  prep_buf_loc = prep_buf;
  send_buf_loc = send_buf;

  while ((repeat_tag_loc = strstr(send_buf_loc, repeat_tag)))
    {
    tags_processed = 1;

    /* check for nesting tags */
    while ((repeat_tag_nest_loc = strstr(repeat_tag_loc + strlen(repeat_tag),
        repeat_tag)))
      {
      repeat_tag_loc = repeat_tag_nest_loc;
      nesting = 1;
      }

    /* add text up to tag */
    memcpy(prep_buf_loc, send_buf_loc, repeat_tag_loc - send_buf_loc);

    /* advance prep_buf_loc to end */
    prep_buf_loc += (repeat_tag_loc - send_buf_loc);

    /* advance send_buf_loc to end of tag */
    send_buf_loc = repeat_tag_loc + strlen(repeat_tag);

    /* parse number of repeats */
    sscanf(send_buf_loc, "%d %d", &repeat_min, &repeat_max);
    if (repeat_max < 1 || repeat_min < 1 || repeat_min > repeat_max)
      return send_buf_len;

    /* location of repeat text */
    if (!(repeat_text_loc = strstr(send_buf_loc, ">")))
      return send_buf_len;
    repeat_text_loc++;

    /* advance send_buf_loc to end of tag data */
    if (!(send_buf_loc = strstr(send_buf_loc, repeat_tag_end)))
      return send_buf_len;

    /* check that location of repeat text didnt fail */
    if (repeat_text_loc > send_buf_loc)
      return send_buf_len;

    /* parse tag */
    repeat_text_len = send_buf_loc - repeat_text_loc;
    memcpy(repeat_text, repeat_text_loc, repeat_text_len);

    /* choose random length */
    int rand_repeats = rand() % ((1 + repeat_max - repeat_min)) + repeat_min;

    /* advance send_buf_loc to end of end_tag name */
    send_buf_loc += strlen(repeat_tag_end);

    /* start repetition */
    int i;
    for (i = 1; i <= rand_repeats; i++)
      {
      memcpy(prep_buf_loc, repeat_text, repeat_text_len);
      prep_buf_loc += repeat_text_len;
      }
    }

  if (tags_processed)
    {
    /* add last part */
    memcpy(prep_buf_loc, send_buf_loc, (send_buf + send_buf_len) - send_buf_loc);

    /* copy new buffer to send buffer */
    memset(send_buf, 0, SESSION_MAX_LEN);
    send_buf_len = (prep_buf_loc - prep_buf + (send_buf + send_buf_len)
        - send_buf_loc);
    memcpy(send_buf, prep_buf, send_buf_len);
    }
  if (nesting == 1)
    goto start;
  else
    return send_buf_len;
  }

int instance_num_data(char *send_buf, struct tcpjunk_data *instance_data,
    int send_buf_len)
  {
  char *instance_tag = "<instance>";
  char prep_buf[SESSION_MAX_LEN];
  char *prep_buf_loc = prep_buf;
  char *send_buf_loc = send_buf;
  char *instance_tag_loc = NULL;
  int tags_processed = 0;
  char instance_str[8];

  memset(prep_buf, 0, sizeof(prep_buf));

  while ((instance_tag_loc = strstr(send_buf_loc, instance_tag)))
    {
    tags_processed = 1;
    /* add text up to tag */
    memcpy(prep_buf_loc, send_buf_loc, instance_tag_loc - send_buf_loc);

    /* advance prep_buf_loc to end */
    prep_buf_loc += (instance_tag_loc - send_buf_loc);

    /* advance send_buf_loc to end of tag */
    send_buf_loc = instance_tag_loc + strlen(instance_tag);

    /* insert data */
    sprintf(instance_str, "%d", instance_data->child_num);
    memcpy(prep_buf_loc, instance_str, strlen(instance_str));
    prep_buf_loc += strlen(instance_str);
    }

  if (tags_processed)
    {
    /* add last part */
    memcpy(prep_buf_loc, send_buf_loc, (send_buf + send_buf_len) - send_buf_loc);

    /* copy new buffer to send buffer */
    memset(send_buf, 0, SESSION_MAX_LEN);
    send_buf_len = (prep_buf_loc - prep_buf + (send_buf + send_buf_len)
        - send_buf_loc);
    memcpy(send_buf, prep_buf, send_buf_len);
    }
  return send_buf_len;
  }

int char_data(char *send_buf, struct tcpjunk_data *instance_data,
    int send_buf_len)
  {
  char *char_tag = "<char ";
  char prep_buf[SESSION_MAX_LEN];
  char *prep_buf_loc = prep_buf;
  char *send_buf_loc = send_buf;
  char *char_tag_loc = NULL;
  int tags_processed = 0;
  int char_hex;

  memset(prep_buf, 0, SESSION_MAX_LEN);

  while ((char_tag_loc = strstr(send_buf_loc, char_tag)))
    {
    tags_processed = 1;
    /* add text up to tag */
    memcpy(prep_buf_loc, send_buf_loc, char_tag_loc - send_buf_loc);

    /* advance prep_buf_loc to end */
    prep_buf_loc += (char_tag_loc - send_buf_loc);

    /* advance send_buf_loc to end of tag */
    send_buf_loc = char_tag_loc + strlen(char_tag);

    /* get char */
    sscanf(send_buf_loc, "%x", &char_hex);

    /* insert character */
    *prep_buf_loc = (unsigned char) char_hex;
    prep_buf_loc++;

    /* advance send_buf_loc to end of tag */
    if (!(send_buf_loc = strstr(send_buf_loc, ">")))
      return send_buf_len;;
    send_buf_loc++;
    }

  if (tags_processed)
    {
    /* add last part */
    memcpy(prep_buf_loc, send_buf_loc, (send_buf + send_buf_len) - send_buf_loc);

    /* copy new buffer to send buffer */
    memset(send_buf, 0, SESSION_MAX_LEN);
    send_buf_len = (prep_buf_loc - prep_buf + (send_buf + send_buf_len)
        - send_buf_loc);
    memcpy(send_buf, prep_buf, send_buf_len);
    }
  return send_buf_len;
  }

